///////////////////////////////////////// --o-
// threads.js
//
// MESSAGES IN--
//   experiment, usage, 
//   dumpglobal, dumpdata, dumpsettings, dumpstate
//
// MESSAGES OUT--
//
// OBJECTS--
//   segmentObject()
//
// PUBLIC FUNCTIONS--
//   experiment(), anything(), dumpglobal()
//   segmentObject :: 
//     dumpdata(), dumpsettings(), setDefaultValues(),
//     incite(), stopExecutionNow(), startExecution()
//
// VERSIONS---
//    0.1: 3 independent voices; external hooks
// 					Mon Jul 20 03:51:04 PDT 2009
//					dreeder@danaprajna.com


//----------------------------------- -o-
SCRIPTNAME = "chorus.js";
VERSION = 0.1

autowatch = 1;		// recompile when editor saves file 
debug = 1;              // global debug: higher value for greater verbosity

inlets = 1;
setinletassist(0, "(message)");


//----------------------------------- -o-
function anything()
{
  var debugstr 	= ds = "anything";
  var args 	= new String( arrayfromargs(arguments) ).split(COMMA);
  var instr 	= "inlet(" + inlet + ")  " + messagename + ": " + args;

  switch (args[0]) 
  {
  default:
    if (debug) {
      db(ds);
      dx(ds, instr);
      de(ds);
      // da(ds, "jsargsuments", jsarguments);
    }

    /*FALL THROUGH*/
  }

  return;
}  // anything()



//----------------------------------- -o-
function dumpglobal()   // [message]
{ 		

  pl(); pl(); pl(); pl(); pl();
  psx("                                   STATE: segment values");

  lo.dumpdata(); 	// line
  lo.dumpsettings();
  lo.dumpstate();

  pl(); pl(); pl(); pl(); pl();
  ps();

}  // dumpglobal()


//----------------------------------- -o-
function usage() { 	// [message]

  pl(); pl(); pl(); pl(); pl();
  psx("                                   METHODS: sequenceObject()");

  infomsg( "incite (  [max_iterations]  )");
  infomsg( "startExecution (  [iter_num],  [iter_max],  [seg_type],  [seg_index]   )");
  infomsg( "stopExecutionNow (  [seg_type],  [seg_index]  )");

  ps();
  infomsgl( "getTypeName()", "string for TYPE_*");

  ps();
  pl(); pl(); pl(); pl(); pl();
  psx("                                   MESSAGES: global scope");
  infomsg( "dumpglobal" + QCOL2 + "state and status for all segment objects");
  infomsg( "     +  all object values  (dumpdata)");
  infomsg( "     +  character of segment object  (dumpsettings)");
  infomsg( "     +  status of drawing and segment iterations (dumpstate)");
  infomsg( "usage" + QCOL2 + "this here informational message");

  pl(); pl(); pl(); pl(); pl();
  ps();
}  // usage()



//----------------------------------- -o-
g = new Global("threads");
g.initTime = new Date(); 		// XXX  vs time() method?

// global contants
//
g.NOTYPE 	= 0;
g.TYPE_LINE 	= 1;
g.TYPE_SPOT 	= 2;
g.TYPE_CURVE 	= 3;
g.MAXTYPE 	= 4;


// segment objects organized by type and index
//
g.segArray = new Array( g.NOTYPE, g.TYPE_LINE, g.TYPE_SPOT, g.TYPE_CURVE );
g.segArray[g.NOTYPE] = undefined;

g.lineArray = g.segArray[g.TYPE_LINE] = new Array();
g.lineArray[0] = undefined;
g.lineArray[1] = 
	new segmentObject(g.TYPE_LINE, 1, "-- 1 --", lineparams, "m_line");

g.spotArray = g.segArray[g.TYPE_SPOT] = new Array();
g.spotArray[0] = undefined;
g.spotArray[1] = 
	new segmentObject(g.TYPE_SPOT, 1, ".. 1 ..", spotparams, "m_spot");

g.curveArray = g.segArray[g.TYPE_CURVE] = new Array();
g.curveArray[0] = undefined;
g.curveArray[1] = 
	new segmentObject(g.TYPE_CURVE, 1, "> 1 <", curveparams, "m_curve");


lo = g.lineArray[1]; 	// XXX  handy for testing...
so = g.spotArray[1]; 	
co = g.curveArray[1]; 



//----------------------------------- -o-
// segmentObject()
//
//   Objects may be stopped and restarted.
//
// ARGUMENTS--
//   _seg_type 		= g.TYPE_*
//   _seg_type_index 	= index into g.<type>Array
//   _seg_name 		= cute string for debugging purposes
//   _seg_param_func 	= pointer function for data manipulation
//   _max_recvtoken 	= name of token argument for Max receive object 
//
// MESSAGES OUT-- 		maxRecvToken
// OUTLETS-- 			n/a
// PUBLIC/PRIVATE FUNCTIONS-- 	Cf. usage()
// CALLS FUNCTIONS-- 		drawsegment(), maxParamFunction
//
segmentObject.local = 1;

function segmentObject(_seg_type, _seg_type_index, _seg_name, 
					_seg_param_func, _max_recvtoken)
{
  //
  // ARGUMENT CHECKS
  //
  if ( (undefined == _seg_type) 
	  || (undefined == _seg_type_index)
	    || (undefined == _seg_name) 
	      || (undefined == _seg_param_func) 
		|| (undefined == _max_recvtoken) ) 
  {
    errmsg( "USAGE: segmentObject("
		+ " <seg_type> <seg_type_index> <seg_name> "
		  + "<seg_param_func> <max_recvtoken> )" );
    return;
  }


  //----------------------------------- -o-
  // object arguments
  //
  this.type 		= _seg_type;
  this.typeIndex 	= _seg_type_index;
  this.name 		= _seg_name;
  this.paramFunction 	= _seg_param_func;
  this.maxRecvToken 	= _max_recvtoken;


  // object task and runtime logic
  //
  this.task = new Task( 
    		drawsegment, this, new Array(this.type, this.typeIndex) );

  this.iterationNum 	= 0; 		// DEFAULT  
  this.iterationMax 	= 0; 
    	// stopping condition for drawing iterations
	// >0: finite iteration count
	//  0: do nothing
	// <0: run (forever) until otherwise stopped 

  this.active 		= false;
    	// true: changes are audibly drawn through time and frequency
  this.activeComplete 	= false;
    	// true: moment audible directive has completed
  this.stopExecution 	= false;
    	// true: stop iterations as soon possible

  this.isRunning 	= false;
  	// true: segment iterations are occuring
  this.isFinished 	= false;
  	// true: segment halts naturally 
	// TBD  handle continuity: 1) across drawing iterations; 2)
	//        between start/stops.
	// FIX  can running or finished be reset automatically?  or
	//        only in scope of object creation?


  //----------------------------------- -o-
  // setDefaultValues( _everything )
  //
  //   Reset dynamic object values to original state at creation.
  //   ASSUMES  object is not running
  //
  //  XXX  be sure all new variables are accounted for...
  //
  this.setDefaultValues = function( _everything )
  {
    var debugstr = ds = "segmentObject::setDefaultValues";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }


    //----------------------------------- -o-
    // state values that are ALWAYS reset to default
    //

    // object runtime logic
    //
    this.active 	= false;
    this.activeComplete = false;
    this.stopExecution 	= false;
    this.isRunning 	= false;
    this.isFinished 	= false;

    // if (debug) { de(ds); }
  }  // segmentObject::setDefaultValues()


  //----------------------------------- -o-
  // dumpdata() 
  //
  this.dumpdata = function()     
  {
    var debugstr = ds = "segmentObject::dumpdata";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }
    
    var msg = "segObj(" + this.name + "):  ";
    msg += "[" + this.iterationNum + SLASH + this.iterationMax + "]";
    infomsg(msg);

    // if (debug) { de(ds); }
  }  // segmentObject::dumpdata()


  //----------------------------------- -o-
  // dumpsettings()
  //
  this.dumpsettings = function()     
  {
    var debugstr = ds = "segmentObject::dumpsettings";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }

    var msg = "segObj(" + this.name + "):  ";
    msg += "[" + this.iterationNum + SLASH + this.iterationMax + "]";
    msg += "     TYP/RCV: " + this.getTypeName() + QSLASH + this.maxRecvToken;
    msg += QDASH + "OPEN/BEGIN: "
		  + this.gateOpen + QSLASH
		    + this.newBeginValues;
    infomsg(msg);

    // if (debug) { de(ds); }
  }  // segmentObject::dumpsettings()
  

  //----------------------------------- -o-
  // dumpstate() 
  //
  this.dumpstate = function()     
  {
    var debugstr = ds = "segmentObject::dumpstate";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }

    var msg = "segObj(" + this.name + "):  ";
    msg += "[" + this.iterationNum + SLASH + this.iterationMax + "]";
    msg += "     A/C, STOP: " 
		  + this.active + QSLASH
		    + this.activeComplete + COMMA + SP
		      + this.stopExecution;
    msg += QDASH + "RUN/FIN: " + this.isRunning + QSLASH + this.isFinished;
    infomsg(msg);

    // if (debug) { de(ds); }
  }  // segmentObject::dumpstate()



  //----------------------------------- -o-
  // isDone
  //
  //   RETURNS true if iteration should stop; else false.
  //   Never returns true if iterationMax < 0.
  //   Stopping conditions:
  //     this.stopExecution is true
  //     iterationNum >= iterationMax
  //
  // XXX  set this.isRunning <-- true from calling environment...
  //
  this.isDone = function()     
  {
    var rval;

    this.iterationNum += 1;
    	// XXX  isDone() should not change the data it observes...

    if (true == this.stopExecution) {
      this.stopExecution = false;
      	// FIX specific to segment iterations?  (ie, not draw iterations)
      this.isRunning = false;
      rval = true;

      if (debug) { 
	infomsg( "segmentObject(" + this.name + "):  "
			+ this.getTypeName() + " HAS BEEN STOPPED." );
      }

    } else if (this.iterationMax < 0) { 	// run forever
      rval = false;
    
    } else if (this.iterationNum <= this.iterationMax) {
      rval = false;

    } else {
      if (debug) {
        infomsg("segmentObject(" + this.name + "):  " 
			+ this.getTypeName() + " IS FINISHED.");
      }

      this.isRunning = false;
      this.isFinished = true;
      rval = true;
    }

    
    return rval; 
  }  // segmentObject::isDone()


  this.stopExecutionNow = function( _seg_type, _seg_index ) 	// -o-
  {
    var debugstr = ds = "segmentObject::stopExecutionNow";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }


    // by default: stop self.
    // XXX  could stop anybody... 
    //
    var tp = (undefined != _seg_type) ? _seg_type : this.type;
    var ip = (undefined != _seg_index) ? _seg_index : this.typeIndex;
    
    var obj = g.segArray[ tp ][ ip ];
    obj.stopExecution = true;


    // if (debug) { de(ds); }
  }  // segmentObject::stopExecutionNow


  //----------------------------------- -o-
  // startExecution()
  //
  // OPTIONALY name segment type/index.
  // OPTIONALY set iterationNum and iterationMax.
  //
  // IF (iterationNum < 0) THEN leave it be (pick up from previous stop)
  // IF (itearationMax < 0) THEN run forever.
  //
  // XXX  there may be dragons.
  // TBD  be able to stop and start a segment sequence at will
  //
  this.startExecution = 
  	function( _iter_num, _iter_max, _seg_type, _seg_index )	// -o-
  {
    var debugstr = ds = "segmentObject::startExecution";
    var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    if (debug) { db(ds); dal(ds, args, "args"); }


    // by DEFAULT: stop self.
    // XXX  could stop anybody... 
    //
    var tp = (undefined != _seg_index) ? _seg_index : this.type;
    var ip = (undefined != _seg_index) ? _seg_index : this.typeIndex;
    
    var obj = g.segArray[ tp ][ ip ];


    // reset iterationNum  -ELSE-  leave current value
    //
    if ( (undefined != _iter_num) && (_iter_num >= 0) ) {
      obj.iterationNum = _iter_num;
    }

    // reset iterationMax  -ELSE-  leave set to inifinity
    //
    if ( (undefined != _iter_max) && (_iter_max >= 0) ) {
      obj.iterationMax = _iter_max;
    }

    obj.stopExecution = false;
    obj.isRunning = true;

    obj.task.execute();


    if (debug) { de(ds); }
  }  // segmentObject::startExecution


  this.incite = function( _max_iterations ) 			// -o-
  {
    var debugstr = ds = "segmentObject::incite";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }


    if (undefined != _max_iterations) {
      this.setIterationBoundaries(0, _max_iterations);
    }

    this.isFinished = false; 	// FIX  rely on setDefaultValues()?
    this.isRunning = false; 	// FIX  rely on setDefaultValues()?

    this.task.execute();


    // if (debug) { de(ds); }
  }  // segmentObject::incite()


  this.setIterationBoundaries = function( _iter_num, _iter_max ) // -o-
  	// XXX  begins counting at (_iter_num + 1)
  {
    var debugstr = ds = "segmentObject::setIterationBoundaries";
    // var args 	 = new String( arrayfromargs(arguments) ).split(COMMA);
    // if (debug) { db(ds); dal(ds, args, "args"); }


    if (undefined != _iter_max) { 	// negative: run forever
      this.iterationMax = _iter_max;
    }

    if ((undefined != _iter_num) && (_iter_num >= 0)) {
      this.iterationNum = _iter_num;
    }


    // if (debug) { de(ds); }
  }  // segmentObject::setIterationBoundaries();
  

  this.getTypeName = function() 
  {
    var type_name = "";

    switch (this.type) { 
    case g.TYPE_LINE: 	type_name = "LINE"; 		break;
    case g.TYPE_SPOT: 	type_name = "SPOT"; 		break;
    case g.TYPE_CURVE: 	type_name = "CURVE"; 		break;
    default: 		type_name = "UNKNOWN"; 		break;
    }

    return (type_name);
  }  // segmentObject::getTypeName

}  // segmentObject()



//----------------------------------- -o-
// drawsegment()   [segmentObject: task function]
//
// INPUT--
//   args[0]: <calling_obj>.type
//   args[1]: <calling_obj>.typeIndex
//
// CALLED BY-- instantiation of a segmentObject()
//
// CALLS METHODS--
//   <calling_obj> :: sendFAE(), isDone()
//
// USES GLOBAL VARIABLES-- g.segArray[]
//
// USES OBJECT VARIABLES--  
//   segmentObject ::
//       type, typeIndex
//       active, activeComplete, paramFunction
//       task.schedule()
//
drawsegment.local = 1;

function drawsegment()
{
  var debugstr 	= ds = "drawsegment";
  var args 	= new String( arrayfromargs(arguments) ).split(COMMA);
  // if (debug) { db(ds); dal(ds, args, "args"); }

  
  // handle arguments
  //	
  var so = g.segArray[ parseInt(args[0]) ][ parseInt(args[1]) ];
  
  if (undefined == so) { 			// so: segment_object
    errmsg( ds + "(): segment_object is undefined." ); 		
    errmsg("USAGE: drawsegment( seg.type, seg.index )");
    return;
  }


  // segment prep, draw, reset...
  //
  if (false == so.active) { 			//*** CONFIGURE
    if (debug) { 
      dx(ds, QSPACE + QSPACE 
      		+ "-------CONFIGURE  (" + so.getTypeName() + ")-------"); 
    }

    if (so.isDone()) { 
      return; 
    }

    // !!!  (re)configures object behavior
    //
    so.paramFunction(so.type, so.typeIndex); 	

    so.active 		= true;
    so.isRunning 	= true; 	
    	// XXX  true ONLY after first call to paramFunction()

  } else if (false == so.activeComplete) { 	//*** DRAW
    if (debug) { 
      dx(ds, QSPACE + QSPACE
      		+ ".......draw  (" + so.getTypeName() + ")......."); 
    }

    so.activeComplete 	= true;
    so.durationTotal 	= so.envelopeDur - so.envelopeRamp;

  } else { 					//*** GATE DIRECTIVE
    if (debug) { 
      dx(ds, QSPACE + QSPACE
      		+ ".......gate  (" + so.getTypeName() + ")......."); 
    }

    so.active 		= false;
    so.activeComplete 	= false;
    so.durationTotal 	= so.silenceBetween;
  }


  this.task.schedule(1 + so.durationTotal);

  // if (debug) { de(ds); }
}  // drawsegment()


//----------------------------------- -o-
// lineparams(_array_index)
//
//   Configure segmentObject(s) to draw lines.
//
lineparams.local = 1;

function lineparams(_obj_type, _obj_index)
{
  var debugstr 	= ds = "lineparams";
  // var args 	= new String( arrayfromargs(arguments) ).split(COMMA);
  // if (debug) { db(ds); dal(ds, args, "args"); }


  // define self
  //
  var o = g.segArray[ _obj_type ][ _obj_index ];

  if (undefined == o) {
    errmsg( "USAGE:  " + ds + "( <obj_type> <obj_index> )" ); 	return;

  } else if (g.TYPE_LINE != o.type) {
    errmsg( ds + "(): object is not TYPE_LINE"); 	return;
  }


  // parameter initialization before all iterations
  //
  if (false == o.active)  { /*EMPTY*/ }


  // parameter changes per iteration
  //

  // if (debug) { de(ds); }
}  // lineparams()


//----------------------------------- -o-
// experiment()
//
// Something risque.
//

function experiment()
{
  var debugstr 	= ds = "experiment";
  // var args 	= new String( arrayfromargs(arguments) ).split(COMMA);
  // if (debug) { db(ds); da(ds, args, "args"); }

  pl(); pl(); pl(); pl(); pl();
  psx("                                   experiment()  BEGIN");
  dhz();


  pl(); pl(); pl(); pl(); pl();
  psx("                                   experiment()  END");

  // if (debug) { de(ds); }
}  // experiment()

